// src/list-markdown.ts
var listMarkdownPlugin = (md, _options) => {
  md.core.ruler.after("inline", "bangle-list-kind-attrs", (state) => {
    const tokens = state.tokens;
    let currentListKind = null;
    for (let i = 0; i < tokens.length; i++) {
      const token = tokens[i];
      if (isBulletListOpen(token)) {
        currentListKind = "bullet";
        token?.attrSet("data-bangle-list-kind", "bullet");
      } else if (isOrderedListOpen(token)) {
        currentListKind = "ordered";
        token?.attrSet("data-bangle-list-kind", "ordered");
      } else if (token?.type === "bullet_list_close" || token?.type === "ordered_list_close") {
        currentListKind = null;
      } else if (token?.type === "list_item_open" && currentListKind) {
        if (!token.attrGet("data-bangle-list-kind")) {
          token.attrSet("data-bangle-list-kind", currentListKind);
        }
      }
    }
    return false;
  });
  md.core.ruler.after("inline", "bangle-task-lists", (state) => {
    const tokens = state.tokens;
    for (let i = 2; i < tokens.length; i++) {
      if (isTodoItem(tokens, i)) {
        convertToTaskItem(tokens, i);
      }
    }
    return false;
  });
  const originalRenderToken = md.renderer.renderToken.bind(md.renderer);
  md.renderer.renderToken = (tokens, idx, options) => {
    const token = tokens[idx];
    if (token?.type === "list_item_open") {
      const kindAttr = token?.attrGet("data-bangle-list-kind");
      if (kindAttr === "task") {
        const checkedAttr = token.attrGet("data-bangle-task-checked") || "false";
        token.attrSet("data-bangle-task-checked", checkedAttr);
      }
    }
    return originalRenderToken(tokens, idx, options);
  };
  md.core.ruler.after(
    "bangle-task-lists",
    "bangle-ordered-list-order",
    (state) => {
      const tokens = state.tokens;
      const orderStack = [];
      for (const token of tokens) {
        switch (token.type) {
          case "ordered_list_open": {
            let start;
            const startAttr = token.attrGet("start");
            if (startAttr != null) {
              start = Number.parseInt(startAttr, 10);
            } else if (token.markup) {
              const match = token.markup.match(/^(\d+)/);
              start = match ? Number.parseInt(match[1], 10) : 1;
            } else {
              start = 1;
            }
            orderStack.push(start);
            token.attrSet("data-bangle-list-kind", "ordered");
            break;
          }
          case "ordered_list_close": {
            orderStack.pop();
            break;
          }
          case "list_item_open": {
            const currentOrder = orderStack[orderStack.length - 1];
            if (orderStack.length > 0 && currentOrder !== void 0) {
              token.attrSet("data-bangle-list-order", String(currentOrder));
              orderStack[orderStack.length - 1] = currentOrder + 1;
            }
            break;
          }
          default:
            break;
        }
      }
      return false;
    }
  );
};
function isOrderedListOpen(token) {
  return token?.type === "ordered_list_open";
}
function isBulletListOpen(token) {
  return token?.type === "bullet_list_open";
}
function isTodoItem(tokens, index) {
  return isInline(tokens[index]) && isParagraphOpen(tokens[index - 1]) && isListItemOpen(tokens[index - 2]) && startsWithTodoMarkdown(tokens[index]?.content);
}
function startsWithTodoMarkdown(content) {
  if (!content) return false;
  const prefix = content.slice(0, 4).toLowerCase();
  return prefix === "[ ] " || prefix === "[x] ";
}
function convertToTaskItem(tokens, index) {
  const listItemOpen = tokens[index - 2];
  const inlineToken = tokens[index];
  if (!listItemOpen || !inlineToken || !inlineToken.children) return;
  listItemOpen.attrSet("data-bangle-list-kind", "task");
  const firstChild = inlineToken.children[0];
  if (firstChild?.type === "text") {
    const text = firstChild.content;
    const isChecked = text[1]?.toLowerCase() === "x";
    listItemOpen.attrSet(
      "data-bangle-task-checked",
      isChecked ? "true" : "false"
    );
    if (text.startsWith("[ ] ") || text.startsWith("[x] ") || text.startsWith("[X] ")) {
      firstChild.content = text.slice(4);
    }
  }
}
function isInline(token) {
  return token?.type === "inline";
}
function isParagraphOpen(token) {
  return token?.type === "paragraph_open";
}
function isListItemOpen(token) {
  return token?.type === "list_item_open";
}

export {
  listMarkdownPlugin
};
